/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 1.3.40
 *
 * Do not make changes to this file unless you know what you are doing--modify
 * the SWIG interface file instead.
 * ----------------------------------------------------------------------------- */

package org.opalvoip;

public class OpalParamCallCleared {
  private long swigCPtr;
  protected boolean swigCMemOwn;

  protected OpalParamCallCleared(long cPtr, boolean cMemoryOwn) {
    swigCMemOwn = cMemoryOwn;
    swigCPtr = cPtr;
  }

  protected static long getCPtr(OpalParamCallCleared obj) {
    return (obj == null) ? 0 : obj.swigCPtr;
  }

  protected void finalize() {
    delete();
  }

  public synchronized void delete() {
    if (swigCPtr != 0) {
      if (swigCMemOwn) {
        swigCMemOwn = false;
        OPALJNI.delete_OpalParamCallCleared(swigCPtr);
      }
      swigCPtr = 0;
    }
  }

  public void setM_callToken(String value) {
    OPALJNI.OpalParamCallCleared_m_callToken_set(swigCPtr, this, value);
  }

  public String getM_callToken() {
    return OPALJNI.OpalParamCallCleared_m_callToken_get(swigCPtr, this);
  }

  public void setM_reason(OpalCallEndReason value) {
    OPALJNI.OpalParamCallCleared_m_reason_set(swigCPtr, this, value.swigValue());
  }

  public OpalCallEndReason getM_reason() {
    return OpalCallEndReason.swigToEnum(OPALJNI.OpalParamCallCleared_m_reason_get(swigCPtr, this));
  }

  public OpalParamCallCleared() {
    this(OPALJNI.new_OpalParamCallCleared(), true);
  }

}
